// (C) 2003 Ulrich Hecht <uli@emulinks.de>
// Licensed under the terms of the GNU Public License version 2

#include <stdio.h>
#include <assert.h>
#include <string.h>

#include "f2a.h"
#include "parport.h"
#include <unistd.h>
#include <stdlib.h>

#define MODE_WRITE 0
#define MODE_READ 1
#define MODE_ERASE 2




int verbose=1;


const char * get_mode_str(int mode){

	switch (mode) {
		case MODE_WRITE: return "MODE_WRITE";
		case MODE_READ: return "MODE_READ" ;
		case MODE_ERASE: return "MODE_ERASE";
		default: return "UNKOWN";
	}
	return NULL;
}

int main(int argc, char** argv)
{
	recvmsg rm;
	int opt;
	int mode=MODE_WRITE;	// default is to write to cart
	int sram=0;				// default is to access ROM, not SRAM
	int size=8;				// 8 MBit/kBit
	int offset=0;			// offset in MBit (ROM) or kBit (SRAM)
	int logo = 0;
	DEVICE_T device = DEVICE_USB;

	assert(sizeof(sendmsg)==63);
	assert(sizeof(recvmsg)==64);

	do
	{
		opt=getopt(argc, argv, "rwbs:o:vd:le");
		switch(opt)
		{
		case 'r':
			mode=MODE_READ;
			break;
		case 'w':
			mode=MODE_WRITE;
			break;
		case 'b':
			sram=1;
			break;
		case 's':
			size=atoi(optarg);
			break;
		case 'o':
			offset=atoi(optarg);
			break;
		case 'v':
			verbose++;
			parport_enable_debug();
			break;
		case 'l':
			logo=1;
			break;
		case 'e':
			mode=MODE_ERASE;
			break;
		case 'd':
			if(!strncmp(optarg,"usb",3))
				device = DEVICE_USB;
			else if (!strncmp(optarg,"parport",7))
				device = DEVICE_PARPORT;
			else
				fprintf(stderr, "Unknow Device: '%s'\n", optarg);
			break;
		case '?':
			fprintf(stderr, "f2a " F2A_VERSION "\n");
			fprintf(stderr, "Usage: %s [OPTION]... [FILE]...\n\n",argv[0]);
			fprintf(stderr, "  -r\tread from cart\n");
			fprintf(stderr, "  -w\twrite to cart (default)\n");
			fprintf(stderr, "  -b\taccess SRAM (default is to access ROM)\n");
			fprintf(stderr, "  -s\tamount of data to read (MBit for ROM, kBit for SRAM)\n");
			fprintf(stderr, "  -o\toffset on cart (MBit for ROM, kBit for SRAM)\n");
			fprintf(stderr, "  -d\tselect f2a device [usb|parport]\n");
			fprintf(stderr, "  -v\tincrease verbosity\n");
			fprintf(stderr, "\n");
			fprintf(stderr, "  -l\ttransfer iLinker Logo\n");
			fprintf(stderr, "  -e\terase cart\n");

			exit(1);
		}
	} while(opt!=-1);

	printf("DEBUG: argc %d optind %d mode:%s\n", argc, optind, get_mode_str(mode));

	if(argc-optind<=0 && (MODE_WRITE==mode || MODE_READ==mode) )
	{
		fprintf(stderr, "%s: file name missing\n",argv[0]);
		exit(2);
	}
	if(mode==MODE_READ && argc-optind != 1)
	{
		fprintf(stderr, "%s: cannot write to multiple files\n",argv[0]);
		exit(3);
	}

	fprintf(stderr, "using device '%s'\n", device ? "parport" : "usb" );

	memset(&rm,0,sizeof(rm));


	if (DEVICE_USB==device) {
		if(usb_connect())
		{
			fprintf(stderr, "%s: unable to connect to F2A USB linker\n",argv[0]);
			exit(1);
		}
	} else {
		if (parport_init(0x378)){
			fprintf(stderr, "%s: unable to connect to F2A Parport linker\n",argv[0]);
			exit(1);
		}

	}


	if (DEVICE_USB==device) {
		f2a_info(&rm);
		if(rm.data[0]==0)
		{
			if(verbose) fprintf(stderr, "Please turn on GBA with SELECT and START held down.\n");
			f2a_boot(FIRMWARE_GBA);
			f2a_info(&rm);
		}
	} else {
		if (parport_boot(logo)){
			fprintf(stderr, "Please turn on GBA with SELECT and START held down.\n");
			parport_exit();
			exit(1);
		}
	}
	if (device==DEVICE_USB) {

		if(mode==MODE_WRITE)
		{
			if(sram==0)
				f2a_writemem(argc-optind,argv+optind,0x8000000+offset*128*1024);
			else
				f2a_writemem(argc-optind, argv+optind, 0xe000000+offset*128);
		}
		else
		{
			if(sram==0)
				f2a_readmem(argv[optind], 0x8000000+offset*128*1024, size*128*1024);
			else
				f2a_readmem(argv[optind], 0xe000000+offset*128, size*128);
		}

		usb_disconnect();
	} else {

		if (MODE_WRITE==mode){
			if(sram==0)
				parport_write(argc-optind,argv+optind, 0x8000000+offset);
			else
				parport_write(argc-optind,argv+optind, 0xe000000+offset);
		} else if (MODE_ERASE==mode){
			if(sram==0)
				parport_erase(0x08000000, size*128*1024);
			else
				parport_erase(0xe000000, size*128*1024);
		} else if (MODE_READ) {
			if(sram==0)
				parport_read(0x08000000, size*128*1024,argv[optind]);
			else
				parport_read(0xe000000, size*128*1024,argv[optind]);


		}

	}
	return 0;
}
