/*
Author: David Voswinkel
Email:  d.voswinkel@netcologne 
Date:   2003-07-27 16:43:09

$Id: parport.c,v 1.3 2003/07/27 17:24:10 david Exp $

$Log: parport.c,v $
Revision 1.3  2003/07/27 17:24:10  david
added read cart

Revision 1.2  2003/07/27 14:28:51  david
*** empty log message ***

*/
 

#include <stdio.h>
#include <time.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <signal.h>
#include <unistd.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/io.h>
#include <sys/timeb.h>
#include <sys/time.h>
#include "parport.h"

extern unsigned char img_mbilinker[];
extern unsigned char img_logo[];
extern unsigned char img_loader[];

static int parport_debug = 0;
static int pdata;
static int pstat;
static int pctl;
static int eppaddr;
static int eppdata;
static int ecpreg;
static int eppmode;




static int parport_send_buffer(int cmd,int addr,
							int size,unsigned  char * resource,int head,int flip,unsigned int exec,int mode);

static int parport_send_cmd(int cmd,int addr, int size);
static int parport_exec_cmd(int cmd,int addr, int size);
static int parport_receive_data(int cmd,int addr, int size,unsigned char * filename, int flip);

static int parport_send_head(int cmd,int size);
static int parport_send_raw(unsigned char * buffer, int len);
static int parport_receive_raw(unsigned char * buffer, int len);
static int parport_wait();
static void parport_out31(int val);
static void parport_out91(int val);
static void parport_nop();
static void parport_init_delay();
static unsigned int parport_flip(unsigned int val);
static int delay_loop;


const unsigned char parport_gba_header[] = {
						0xea, 0x00, 0x00, 0x2e, 0x51, 0xae, 0xff, 0x24, 0x21, 0xa2, 0x9a, 0x69, 0x0a, 0x82, 0x84, 0x3d,
                        0xad, 0x09, 0xe4, 0x84, 0x98, 0x8b, 0x24, 0x11, 0x21, 0x7f, 0x81, 0xc0, 0x19, 0xbe, 0x52, 0xa3,
                        0x20, 0xce, 0x09, 0x93, 0x4a, 0x4a, 0x46, 0x10, 0xec, 0x31, 0x27, 0xf8, 0x33, 0xe8, 0xc7, 0x58,
                        0xbf, 0xce, 0xe3, 0x82, 0x94, 0xdf, 0xf4, 0x85, 0xc1, 0x09, 0x4b, 0xce, 0xc0, 0x8a, 0x56, 0x94,
                        0xfc, 0xa7, 0x72, 0x13, 0x73, 0x4d, 0x84, 0x9f, 0x61, 0x9a, 0xca, 0xa3, 0x27, 0xa3, 0x97, 0x58,
                        0x76, 0x98, 0x03, 0xfc, 0x61, 0xc7, 0x1d, 0x23, 0x56, 0xae, 0x04, 0x03, 0x00, 0x84, 0x38, 0xbf,
                        0xfd, 0x0e, 0xa7, 0x40, 0x03, 0xfe, 0x52, 0xff, 0xf1, 0x30, 0x95, 0x6f, 0x85, 0xc0, 0xfb, 0x97,
                        0x25, 0x80, 0xd6, 0x60, 0x03, 0xbe, 0x63, 0xa9, 0xe2, 0x38, 0x4e, 0x01, 0xff, 0x34, 0xa2, 0xf9,
                        0x44, 0x03, 0x3e, 0xbb, 0xcb, 0x90, 0x00, 0x78, 0x94, 0x3a, 0x11, 0x88, 0x63, 0x7c, 0xc0, 0x65,
                        0xaf, 0x3c, 0xf0, 0x87, 0x8b, 0xe4, 0x25, 0xd6, 0x72, 0xac, 0x0a, 0x38, 0x07, 0xf8, 0xd4, 0x21,
                        };



const unsigned char header[] = 	{ 	0x49,0x2d,0x4c,0x69,0x6e,0x6b,0x65,0x72,
										0x2e,0x31,0x30,0x30,0x00,0x00,0x01,0xe8};

void parport_enable_debug()
{
	parport_debug++;
}


int parport_init(int pbase)
{
	int stat;	
	parport_init_delay();
	pdata 	= pbase;    // 0x378
	pstat 	= pbase + 1;// 0x379
	pctl	= pbase + 2;// 0x37a
	eppaddr	= pbase + 3;// 0x37b
	eppdata	= pbase + 4;// 0x37c
	ecpreg	= pbase + 0x402;
	eppmode = 1;

	fprintf (stderr, "Using Base port 0x%-4x\n", pbase);

	if (getuid () && geteuid ()) {
        fprintf (stderr, "ERROR: No privileges to use I/O\n");
       	return -1;
    }

	if(ioperm(pdata, 3, 1)){
        fprintf (stderr, "ERROR: ioperm failed\n");
		return -1;
	}


	if(iopl(3)){
        fprintf (stderr, "ERROR: iopl failed\n");
		return -1;
	}

	if (getgid () && !getegid ())
        setegid (getgid ());

    if (getuid () && !geteuid ())
        seteuid (getuid ());


	outb(0x04, pctl);
	outb(0x01, pstat);
	outb(0x01, pctl);
	outb(0x04, pdata);

	parport_out91(0x47);
	parport_out31(0x02);
	parport_out91(0x12);
	parport_out31(0x01);
	parport_out91(0x34);
	parport_out31(0x00);
	parport_out91(0x56);
	parport_out31(0x02);
	parport_out91(0x00);

	outb(0x02, pctl);
	stat = inb(pstat);	
#if 0	
	if(stat!=0x7e){
        fprintf (stderr, "ERROR: parport init failed (%i) \n",stat);
		return -1;
	}
#endif	
	outb(0x06, pctl);
	stat = inb(pstat);	
#if 0
	if(stat!=0x7e){
        fprintf (stderr, "ERROR: parport init failed (%i)\n",stat);
		return -1;
	}
#endif 

	outb(0x01, pstat);
	outb(0x01, pctl);
	outb(0x04, eppaddr);
	outb(0x07, eppdata);

	outb(0x01, pctl);
	outb(0x02, eppaddr);
	outb(0x12, eppdata);

	outb(0x01, pctl);
	outb(0x01, eppaddr);
	outb(0x34, eppdata);

	outb(0x01, pctl);
	outb(0x00, eppaddr);
	outb(0x56, eppdata);

	outb(0x01, pctl);
	outb(0x01, pctl);
	outb(0x02, eppaddr);
	outb(0x00, pctl);
	if(inb(eppdata)!=0xff){
        fprintf (stderr, "ERROR: epp init failed \n");
		return -1;
	}
	return 0;

}


int parport_boot(int upload_logo)
{
	unsigned char recv[4];
	fprintf(stdout,"Booting GBA...\n");
	if (parport_send_head(PP_HEAD_BOOT,1))
		return -1;
	if (parport_receive_raw(recv,4))
		return -1;
	if (upload_logo){
		if (parport_send_buffer(PP_CMD_WRITEDATA,LOGO_ADDR,LOGO_SIZE,img_logo,0,0,0,0))
			return -1;
	}
	fprintf(stdout,"Uploading iLinker Client.\n");
	if (parport_send_buffer(PP_CMD_WRITEDATA,EXEC_STUB,BOOT_SIZE,img_mbilinker,HEAD,FLIP,EXEC,0)){
		fprintf(stdout,"Upload failed.\n");
		return -1;
	}
	return 0;

}


int parport_write(int files_cnt,char ** files,unsigned int addr)
{
	int j;
	struct stat st;
	int size;

	if (files_cnt>1){
		fprintf(stdout,"Uploading Multiloader.\n");
		if(parport_send_buffer(PP_CMD_WRITEROM,addr,LOADER_SIZE,img_loader,HEAD,FLIP,0,0))
			return -1;
		addr+=LOADER_SIZE;
	}
	for(j=0; j<files_cnt; j++){
		if(stat(files[j],&st)==-1){
			perror("failed to stat file");
			return -1;
		}
		size=st.st_size;
		if(size & (32768-1)) size+=32768;
		size&=~(32768-1);
		fprintf(stdout,"Uploading %s %i Kbytes, padded to %i Kbytes.\n", files[j],(int)(st.st_size/1024),(int)(size/1024));
		if (parport_send_buffer(PP_CMD_WRITEROM,addr,size,files[j],HEAD,FLIP,0,1))
			return -1;

		addr+=size;
	}
	return 0;
}

int parport_erase(unsigned int start,unsigned int size )
{
	int end,addr;
	parport_exec_cmd(PP_CMD_READDATA,ERASE_STUB,1024);
	end = start + (size);

	fprintf(stdout,"Erase Cart start=0x%08x end=0x%08x \n",start,end);
	for (addr=start; addr < end ;addr+=0x40000){
		parport_send_cmd(PP_CMD_ERASE,addr,1024);
	}
	return 0;
}


int parport_read(unsigned int start,unsigned int size, unsigned char * filename )
{
	parport_exec_cmd(PP_CMD_READDATA,ERASE_STUB,1024);
	fprintf(stdout,"Reading from Cart start=0x%08x size=0x%08x \n",start,size);
	parport_receive_data(PP_CMD_READDATA,start,size,filename,FLIP);
	return 0;
}



static int parport_send_head(int cmd, int size)
{
	unsigned char trans[] = {0xa,0x8,0xe,0xc,0x2,0x0,0x6,0x4};
	parport_msg_head_t msg_head;
	memset(&msg_head, 0, sizeof(parport_msg_head_t));
	memcpy(&msg_head.head, header,16);
	msg_head.cmd = cmd;
	msg_head.size = (size/1024)&0xffff;
	msg_head.unkown = (trans[((msg_head.size&255)/32)] << 4 ) | (((1023 - (msg_head.size&1023) ) / 256)&0x0f);
	//msg_head.unkown=0x82;
	if (parport_send_raw((unsigned char*)&msg_head,sizeof(parport_msg_head_t)))
		return -1;
	return 0;
}


static int parport_exec_cmd(int cmd,int addr, int size)
{
	unsigned char *buffer;
	parport_msg_cmd_t msg_cmd;
	memset(&msg_cmd, 0, sizeof(parport_msg_cmd_t));
	msg_cmd.magic = parport_flip(PP_MAGIC_NUMBER);
	msg_cmd.cmd = parport_flip(cmd);
	msg_cmd.addr = parport_flip(addr);
	msg_cmd.sizekb = parport_flip(size/1024);

	msg_cmd.exec_stub = parport_flip(EXEC_STUB);
	msg_cmd.exec = parport_flip(0x08);
	parport_send_head(PP_HEAD_READ,size);
	parport_wait();

	if (parport_debug)
		fprintf (stderr, "sending msg_cmd cmd='0x%08x' addr='0x%08x' size='0x%08x' %i bytes\n",
		msg_cmd.cmd,msg_cmd.addr,msg_cmd.sizekb, sizeof(parport_msg_cmd_t) );


	parport_send_raw((unsigned char*)&msg_cmd,sizeof(parport_msg_cmd_t));
	//parport_wait();
	buffer = (unsigned char *) malloc(size * sizeof(unsigned char));
	parport_receive_raw(buffer,size);;

	return 0;
}

static int parport_receive_data(int cmd,int addr, int size,unsigned char * filename, int flip)
{
	unsigned char recv[4];
	unsigned char buffer[1024];
	unsigned char *mbuffer;

	int romf=-1;
	int i,j,n;
	int *ptr;
	int progress;
	parport_msg_cmd_t msg_cmd;
	memset(&msg_cmd, 0, sizeof(parport_msg_cmd_t));
	msg_cmd.magic = parport_flip(PP_MAGIC_NUMBER);
	msg_cmd.cmd = parport_flip(cmd);
	msg_cmd.addr = parport_flip(addr);
	msg_cmd.sizekb = parport_flip(size/1024);

	if (parport_send_head(PP_HEAD_READ,size))
		return -1;

	if (parport_receive_raw(recv,4))
		return -1;

	if (parport_debug)
		fprintf (stderr, "sending msg_cmd cmd='0x%08x' addr='0x%08x' size='0x%08x' %i bytes\n",
		msg_cmd.cmd,msg_cmd.addr,msg_cmd.sizekb, sizeof(parport_msg_cmd_t) );

	parport_send_raw((unsigned char*)&msg_cmd,sizeof(parport_msg_cmd_t));

	romf=open(filename, O_WRONLY|O_CREAT,0666);


#if 1
	if(romf==-1){
			perror("unable to open rom.");
			return -1;
	}
	mbuffer = (unsigned char *) malloc(size*sizeof(unsigned char));
	parport_receive_raw(mbuffer,size);
	if (flip){
		ptr = (int *) mbuffer;
		for (j=0;j<size/4;j++)
			ptr[j] = parport_flip(ptr[j]);
	}
	//memcpy(buffer,parport_gba_header,HEAD_SIZE);
	if(write(romf, mbuffer, size)==-1){
			perror("unable to write data to file");
			close(romf);
			return -1;
	}
	close(romf);
#else
	for (i=0;i<size;i+=1024){
		parport_receive_raw(buffer,1024);;
		if (flip){
			ptr = (int *) buffer;
			for (j=0;j<256;j++)
				ptr[j] = parport_flip(ptr[j]);
		}

		if(write(romf, buffer, 1024)==-1)
		{
			perror("unable to write data to file");
			close(romf);
			return -1;
		}

		if (parport_debug){
			fprintf (stderr, "reading %i.chunk of %i\n", (int)(i/1024)+1, (int)(size/1024));
		} else {
			progress = (int) ((100.0/size) * (i+(1024)));
			fprintf (stdout,"\r[");
				for (n=0;n<=100;n++)
				if (n<=progress)
					fprintf (stdout,"=");
				else
					fprintf (stdout," ");
			fprintf (stdout,"]\r");
			fflush(stdout);
		}
	}
	if (!parport_debug)
		fprintf (stdout,"\n");
	close(romf);
#endif

	return 0;
}


static int parport_send_cmd(int cmd,int addr,int size)
{
	unsigned char recv[4];
	parport_msg_cmd_t msg_cmd;
	memset(&msg_cmd, 0, sizeof(parport_msg_cmd_t));
	msg_cmd.magic = parport_flip(PP_MAGIC_NUMBER);
	msg_cmd.cmd = parport_flip(cmd);
	msg_cmd.addr = parport_flip(addr);
	msg_cmd.sizekb = parport_flip(size/1024);

	if (parport_send_head(PP_HEAD_WRITE,size))
		return -1;

	if (parport_receive_raw(recv,4))
		return -1;

	if (parport_debug)
		fprintf (stderr, "parport_send_cmd cmd='0x%08x' addr='0x%08x' size='0x%08x' %i bytes\n",
		msg_cmd.cmd,msg_cmd.addr,msg_cmd.sizekb, sizeof(parport_msg_cmd_t) );

	if(parport_send_raw((unsigned char*)&msg_cmd,sizeof(parport_msg_cmd_t)))
		return -1;
	return 0;
}
static int parport_send_buffer(int cmd,int addr,
							int size,unsigned  char *resource,int head,int flip,unsigned int exec,int mode)
{
	unsigned char recv[4];
	int i,j,n;
	int *ptr;
	int progress;
	unsigned char buffer[1024];
	int romf=-1;
	parport_msg_cmd_t msg_cmd;
	memset(&msg_cmd, 0, sizeof(parport_msg_cmd_t));
	msg_cmd.magic = parport_flip(PP_MAGIC_NUMBER);
	msg_cmd.cmd = parport_flip(cmd);
	msg_cmd.addr = parport_flip(addr);
	msg_cmd.sizekb = parport_flip(size/1024);
	if (exec){
		msg_cmd.exec_stub = parport_flip(EXEC_STUB);
		msg_cmd.exec =  parport_flip(0x08);
	}
	if (parport_send_head(PP_HEAD_WRITE,size))
		return -1;
	if (parport_receive_raw(recv,4))
		return -1;

	if (parport_debug)
		fprintf (stderr, "parport_send_buffer cmd='0x%08x' addr='0x%08x' size='0x%08x' %i bytes\n",
		msg_cmd.cmd,msg_cmd.addr,msg_cmd.sizekb, sizeof(parport_msg_cmd_t) );

	if(parport_send_raw((unsigned char*)&msg_cmd,sizeof(parport_msg_cmd_t)))
		return -1;

	if (1==mode){
		romf=open(resource,O_RDONLY);
		if(romf==-1){
			perror("unable to open rom.");
			return -1;
		}
	}

	for (i=0;i<size;i+=1024){
		if (1==mode){
			if(read(romf, buffer, 1024)==-1){
				perror("failed to read from file");
				return -1;
			}
		} else {
			memcpy(buffer,resource,1024);
		}
		if (flip){
			ptr = (int *) buffer;
			for (j=0;j<256;j++)
				ptr[j] = parport_flip(ptr[j]);
		}
		if (!i && head){
			memcpy(buffer,parport_gba_header,HEAD_SIZE);
		}
		if (parport_debug){
			fprintf (stderr, "sending %i.chunk of %i\n", (int)(i/1024)+1, (int)(size/1024));
		} else {
			progress = (int) ((100.0/size) * (i+1024));
			fprintf (stdout,"\r[");
				for (n=0;n<=100;n++)
				if (n<=progress)
					fprintf (stdout,"=");
				else
					fprintf (stdout," ");
			fprintf (stdout,"]\r");
			fflush(stdout);
		}
		parport_send_raw(buffer,1024);
		if (0==mode) resource+=1024;
	}
	if (!parport_debug)
		fprintf (stdout,"\n");
	if (1==mode)
		close(romf);
	return 0;
}

static void parport_dump_byte(unsigned char byte){
	char i;
	for (i=7; i>-1; i--){
		if ((byte>>i)&1)
			fprintf(stderr,"1");
		else
			fprintf(stderr,"0");
	}
	fprintf(stderr,"\n");
}


static int parport_receive_raw(unsigned char * buffer, int len){

	int err;
	int i;
	unsigned char *ptr;
	unsigned char nibble;
	struct timespec t;
	t.tv_sec = 0;
	t.tv_nsec = 1;

	ptr=buffer;
	if (parport_debug)
		fprintf(stderr, "\nreceive:\n%04x: ",0);

	for (err=0,i=0; i<len*2; i++){
		nibble = 0;
		outb(4, pctl);
		parport_nop();
		while ((inb(pstat)&0x80)!=0) nanosleep(&t,NULL);
		outb(5, pctl);
		nibble = inb(pstat);
		parport_nop();
		while ((inb(pstat)&0x80)!=0x80)  nanosleep(&t,NULL);
		if (i%2){
			*ptr |= (nibble>>3)&0x0f;
			if (parport_debug) {
				fprintf(stderr, "%02x ",(unsigned char)*ptr);
				if (!(((i/2)+1)%32) && i && (i/2)<len-1)
					fprintf(stderr, "\n%04x: ",(i/2)+1);
			}
			ptr++;
		} else {
			*ptr |= ((nibble>>3)&0xf)<<4;
		}
	}
	if (parport_debug)
		fprintf(stderr, "\n");

	return err;
}



static int parport_send_raw(unsigned char * buffer, int len)
{
	int timeout;
	int i;
	unsigned char *pc;
	struct timespec t;
	t.tv_sec = 0;
	t.tv_nsec = 1;
	pc = buffer;
	if (parport_debug)
		fprintf(stderr, "\nsend:\n%04x: ",0);
	for (i=0; i<len; i++){
		timeout=2000;
		if (parport_debug) {
			fprintf(stderr, "%02x ",(unsigned char)*pc);
			if (!((i+1)%32) && i && i<len-1)
				fprintf(stderr, "\n%04x: ",i+1);
		}
		outb(0x04, pctl);
		parport_nop();
		while((inb(pstat)&0x80)!=0 && (timeout--)>0) nanosleep(&t,NULL);
		outb(*pc,pdata);
		parport_nop();
  		while((inb(pstat)&0x80)!=0 && (timeout--)>0) nanosleep(&t,NULL);
		outb(0x05, pctl);
		parport_nop();
		while((inb(pstat)&0x80)!=0x80 && (timeout--)>0) nanosleep(&t,NULL);
		pc++;
		if (timeout<0){
			fprintf(stdout,"\nTimeout\n");
			return -1;
		}
	}
	if (parport_debug)
		fprintf(stderr,"\n");

	return 0;
}

static int parport_wait()
{
	int stat;
	while(1){
		outb(0x04, pctl);
		parport_nop();
		stat = inb(pstat);
		if ((stat&0x80)==0x80)
			break;
		outb(0x05, pctl);
		parport_nop();
		inb(pstat);
	}
	return 0;
}

int parport_exit() {

  	if (ioperm(pdata, 3, 0)){
 		fprintf (stderr, "ERROR: ioperm baseport failed\n");
      	return -1;
	}
	return 0;
}

static unsigned int parport_flip(unsigned int val)
{
	return  	((((val&0xff) << 24)&0xff000000) |
				(((val&0xff00) << 8)&0xff0000) |
				(((val&0xff0000) >> 8)& 0xff00) |
				(((val&0xff000000) >> 24)&0xff));

 }

static void parport_init_delay()
{
	struct timeval t0,t1;
	long diff;
	int loop,i,l;
	int target,hit,trials,cnt;
	int hits[10];
	trials = 1000;
	loop = 4000;
	target = 8;
	hit=10;
	diff=0;
	cnt=0;
	while (cnt!=hit) {
		for (l=0;l<trials;l++){
			i = loop;
			gettimeofday(&t0,NULL);
			while(i--);
			gettimeofday(&t1,NULL);
			diff+=(t1.tv_usec - t0.tv_usec);
		}
		diff/=(trials/10);
		loop+=100;
	if (diff>=(target*10)){
			hits[cnt] = loop;
			cnt++;
		}
	}
	loop=0;
	for (i=0;i<hit;i++) loop+=hits[i];
	delay_loop=loop/hit;

	printf("Using Delay Loop %i \n",delay_loop);
}

static void parport_nop()
{
	int i=delay_loop;
	while (i--);
}

static void parport_out31(int val)
{
	outb(0x03, pctl);
	outb(0x01, pctl);
	outb(val, pdata);
}

static void parport_out91(int val)
{
	outb(0x09, pctl);
	outb(0x01, pctl);
	outb(val, pdata);
}
